# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import functools
import inspect
import json
import os
import shutil
import tempfile
from collections import namedtuple
from contextlib import contextmanager
from logging import getLogger

import marv_api as marv  # pylint: disable=unused-import
from marv_cli import create_loglevels
from marv_node import run as marv_node_run
from marv_node.node import Node
from ..run import run_nodes as _run_nodes
from ..setid import SetID

create_loglevels()

marv_node_run.RAISE_IF_UNFINISHED = True

StubDataset = namedtuple('StubDataset', 'setid name collection files time_added timestamp')
StubFile = namedtuple('StubFile', 'idx path missing mtime size')

KEEP = os.environ.get('KEEP')
LOG = getLogger(__name__)


@contextmanager
def temporary_directory(keep=KEEP):
    """Create and cleanup temporary directory."""
    tmpdir = tempfile.mkdtemp()
    LOG.debug('created temporary directory %r', tmpdir)
    try:
        yield tmpdir
    finally:
        if not keep:
            shutil.rmtree(tmpdir)
            LOG.debug('cleaned up temporary directory %r', tmpdir)
        else:
            LOG.debug('keeping temporary directory %r by request', tmpdir)


def make_dataset(files=None, setid=None, name=None, collection=None,
                 time_added=None, timestamp=None):
    # pylint: disable=too-many-arguments
    setid = SetID(42) if setid is None else setid
    name = 'NAME' if name is None else name
    collection = 'COL' if collection is None else collection
    files = [StubFile(i, x, False, 42, os.stat(x).st_size)
             for i, x in enumerate([] if files is None else files)]
    time_added = 0 if time_added is None else time_added
    timestamp = 0 if timestamp is None else timestamp
    return StubDataset(setid=setid, name=name, collection=collection,
                       files=files, time_added=time_added,
                       timestamp=timestamp)


async def run_nodes(dataset, nodes, store=None, persistent=None, **kw):
    """Wrap marv_node.run.run_nodes gathering streams for provided nodes."""
    nodes = [x if isinstance(x, Node) else Node.from_dag_node(x) for x in nodes]
    persistent = {
        k: v if isinstance(v, Node) else Node.from_dag_node(v)
        for k, v in (persistent or {}).items()
    }
    streams = {}
    await _run_nodes(dataset,
                     nodes,
                     {} if store is None else store,
                     persistent=persistent,
                     _gather_into=streams,
                     **kw)
    return [streams.get(x) for x in nodes]


class Spy:  # pylint: disable=too-few-public-methods
    def __init__(self):
        self.requests = []


def make_spy(node):
    spy = Spy()
    requests = spy.requests
    real_invoke = node.invoke

    @functools.wraps(real_invoke)
    def invoke(inputs=None):
        gen = real_invoke(inputs)
        try:
            req = next(gen)
        except StopIteration:
            return

        while True:
            requests.append(req)
            send = yield req
            try:
                req = gen.send(send)
            except StopIteration:
                return

    node.invoke = invoke
    return spy


class TestCase:  # pylint: disable=too-few-public-methods
    BAGS = None
    MARV_TESTING_RECORD = os.environ.get('MARV_TESTING_RECORD')

    def assertNodeOutput(self, output, node):  # pylint: disable=invalid-name
        name = node.__marv_node__.function.rsplit('.', 1)[1]
        testfile = inspect.getmodule(type(self)).__file__
        dirpath = os.path.join(os.path.dirname(testfile), 'output')
        if not os.path.isdir(dirpath):
            os.mkdir(dirpath)
        outfile = os.path.join(dirpath, name + '.json')
        out = [x.to_dict(which=True) for x in output]
        if self.MARV_TESTING_RECORD:
            with open(outfile, 'w') as f:
                json.dump(out, f, sort_keys=True, indent=2)
        else:
            with open(outfile) as f:
                assert json.loads(json.dumps(out)) == json.load(f)



# Copyright 2016 - 2020  Ternaris, all rights reserved.
# SPDX-License-Identifier: AGPL-3.0-only

import re
from collections import namedtuple
from enum import Enum

INVALID_CHARS = tuple(chr(x) for x in range(0x20))
RESERVED_WORDS = (
    'inf', 'infinity', 'Inf', 'Infinity', 'nan', 'NaN', 'True', 'False', 'None',
)
STATE = Enum('ParserState', 'START LIST')
TOKEN_END = re.compile(r'[\s\)]', flags=re.UNICODE)
WHITESPACE = re.compile(r'\s', flags=re.UNICODE)
List = namedtuple('List', 'args start stop')
Literal = namedtuple('Literal', 'value start stop')
Identifier = namedtuple('Identifier', 'name start stop')


class SexpError(Exception):
    """Error scanning s-expression.

    SexpError(error_msg, sexp, idx)
    """


class InvalidCharacter(SexpError):
    """Invalid character within string.

    InvalidCharacter('\x00', sexp, idx)
    """


class InvalidNumber(SexpError):
    """Token is not a valid number but starts like one.

    InvalidNumber('1a', sexp, idx)
    """


class InvalidIdentifier(SexpError):
    """Identifier is invalid.

    InvalidIdentifier('a-b', sexp, idx)
    """


class InvalidStringEscape(SexpError):
    """Invalid string escape sequence.

    InvalidStringEscape('\\uXYZ', sexp, idx)
    """


class UnexpectedCharacter(SexpError):
    """Unexpected character encountered at given index.

    UnexpectedCharacter('+', sexp, idx)
    """


class ReservedWord(SexpError):
    """Identifier is a reserved word.

    ReservedWord('None', sexp, idx)
    """


def _scan_number(sexp, idx):
    start = idx
    string = ''
    while idx < len(sexp):
        char = sexp[idx]
        if TOKEN_END.match(char):
            idx -= 1
            break
        string += char
        idx += 1

    # pylint: disable=too-many-boolean-expressions
    if string[0] == '0' and len(string) > 1 or \
       string == '-0' or \
       string[:2] == '-0' and string[2] != '.' or \
       string[0] == '-' and string[1:] in RESERVED_WORDS:
        raise InvalidNumber(string, sexp, start)

    try:
        return Literal(int(string), start, idx), idx
    except ValueError:
        pass

    try:
        return Literal(float(string), start, idx), idx
    except ValueError:
        raise InvalidNumber(string, sexp, start)


def _scan_string(sexp, idx):
    start = idx
    idx += 1
    string = ''
    while idx < len(sexp):
        char = sexp[idx]
        if char == '"':
            break

        if char == '\\':
            esc, idx = _scan_string_escape(sexp, idx)
            string += esc

        elif char in INVALID_CHARS:
            raise InvalidCharacter(char, sexp, start)

        else:
            string += char

        idx += 1
    else:
        raise SexpError('Runaway string', sexp, start)

    return Literal(string, start, idx), idx


def _scan_string_escape(sexp, idx):
    start = idx
    idx += 1
    char = sexp[idx]
    esc = {
        '"': '"',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        '\\': '\\',
        '/': '/',
    }.get(char)

    if esc:
        return esc, idx

    if char == 'u':
        hexcode = sexp[idx+1:idx+5]
        try:
            if len(hexcode) != 4:
                raise ValueError()
            esc = chr(int(hexcode, 16))
        except ValueError:
            raise InvalidStringEscape(rf'\u{hexcode}', sexp, start)
        return esc, idx + 4

    raise InvalidStringEscape(f'\\{char}', sexp, start)


def _scan_identifier(sexp, idx):
    start = idx
    string = ''
    while idx < len(sexp):
        char = sexp[idx]
        if TOKEN_END.match(char):
            idx -= 1
            break

        string += char
        idx += 1

    if not string.isidentifier():
        raise InvalidIdentifier(string, start, idx)

    if string in RESERVED_WORDS:
        raise ReservedWord(string, start, idx)

    return Identifier(string, start, idx), idx


def scan(sexp):  # noqa: C901  # pylint: disable=too-many-branches
    """Scan sexp into (nested) list of tokens.

    An sexp is a space-separated list of tokens surrounded by
    parentheses. Tokens are lists or json literals.

    (null)
    (0 1 2)
    (true (false))
    (identifier ("foo" 1.2 -1e10 -100 "\u0022 \" \\ \b \f \n \r \t"))

    """
    if sexp[0:1] != '(':
        raise SexpError("Expected '('", sexp, 0)

    stack = [0]
    idx = 1
    while idx < len(sexp):
        char = sexp[idx]

        if char == '(':
            stack.append(idx)

        elif char == ')':
            tokens = []
            item = stack.pop()
            while not isinstance(item, int) and stack:
                tokens.insert(0, item)
                item = stack.pop()
            stack.append(List(tuple(tokens), item, idx))

        elif char == 'n' and sexp[idx:idx+4] == 'null':
            stack.append(Literal(None, idx, idx+3))
            idx += 3

        elif char == 't' and sexp[idx:idx+4] == 'true':
            stack.append(Literal(True, idx, idx+3))
            idx += 3

        elif char == 'f' and sexp[idx:idx+5] == 'false':
            stack.append(Literal(False, idx, idx+4))
            idx += 4

        elif char in '0123456789-':
            number, idx = _scan_number(sexp, idx)
            stack.append(number)

        elif char == '"':
            string, idx = _scan_string(sexp, idx)
            stack.append(string)

        elif char.isalpha():
            identifier, idx = _scan_identifier(sexp, idx)
            stack.append(identifier)

        elif WHITESPACE.match(char):
            pass

        else:
            raise UnexpectedCharacter(char, sexp, idx)

        idx += 1

        if isinstance(stack[-1], int):
            continue

        nextchar = sexp[idx:idx+1]
        if nextchar == ' ' and sexp[idx+1:idx+2] != ')':
            idx += 1

        elif nextchar and not TOKEN_END.match(nextchar):
            raise SexpError('Missing space between tokens', sexp, idx)

    if len(stack) > 1 or not isinstance(stack[0], List):
        raise SexpError('Unbalanced parentheses', sexp, -1)

    return stack[0]



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

# pylint: disable=blacklisted-name,invalid-name

import unittest

from ..node import Node
from ..testing import marv


@marv.node()
@marv.input('a', default=1)
def a_orig(a):  # pylint: disable=unused-argument
    yield


@marv.node()
@marv.input('a', default=1)
def b_orig(a):  # pylint: disable=unused-argument
    yield


class TestCase(unittest.TestCase):
    def test_node_repr(self):
        @marv.node()
        def foo():
            yield
        foo()
        self.assertEqual(
            repr(Node.from_dag_node(foo)),
            '<Node foo.fy4oo6zcym>',
        )

    def test_comparisons(self):
        a = Node.from_dag_node(a_orig)
        b = Node.from_dag_node(b_orig)
        self.assertIs(type(a), type(b))
        self.assertIs(type(a), Node)
        self.assertIs(a, a)
        self.assertIsNot(a, b)

        self.assertLess(a.key, b.key)
        self.assertLess(a, b)
        self.assertLessEqual(a, b)
        self.assertLessEqual(a, a)
        self.assertEqual(a, a)
        self.assertNotEqual(a, b)
        self.assertGreater(b, a)
        self.assertGreaterEqual(b, a)
        self.assertGreaterEqual(b, b)

        self.assertEqual(a, Node.from_dag_node(a_orig.clone()))
        self.assertNotEqual(a, Node.from_dag_node(a_orig.clone(a=2)))



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import datetime
import json
import os
from itertools import count
from logging import getLogger
from pathlib import Path

import mock
import pytest

import marv.model
import marv_api as marv
from marv.app import App
from marv.site import Site
from marv.types import Int8Value, Section, Words
from marv_nodes import dataset as dataset_node


log = getLogger(__name__)


COLLECTION_CONF = """
[collection {name}]
scanner = marv.tests.conftest:scanner
scanroots = /dev/null/{name}
nodes =
    marv_nodes:dataset
    marv_nodes:meta_table
    marv_nodes:summary_keyval
    marv.tests.conftest:node_divisors
{add_nodes}

filters =
    name       | Name       | substring         | string   | (get "dataset.name")
    setid      | Set Id     | startswith        | string   | (get "dataset.id")
    size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
    status     | Status     | any all           | subset   | (status)
    tags       | Tags       | any all           | subset   | (tags)
    comments   | Comments   | substring         | string   | (comments)
    files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
    time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
    time_mtime | Mtime      | lt le eq ne ge gt | datetime | (get "dataset.files[0].mtime")
    divisors   | Divisors   | any all           | subset   | (get "node_divisors.words")
{add_filters}

listing_columns =
    name       | Name     | route    | (detail_route (get "dataset.id") (get "dataset.name"))
    size       | Size     | filesize | (sum (get "dataset.files[:].size"))
    status     | Status   | icon[]   | (status)
    tags       | Tags     | pill[]   | (tags)
    time_added | Added    | datetime | (get "dataset.time_added")
    time_mtime | Mtime    | datetime | (get "dataset.files[0].mtime")
    #divisors   | Divisors | string   | (get "node_divisors.words")
{add_columns}

detail_sections =
{add_sections}
"""


MARV_CONF = """
[marv]
acl = {acl_name}
collections = {collection_names}

{collections}
"""


def recorded(data, path):
    """Assert data corresponds to recorded data.

    Record data if RECORD is set. Return True on success to enable
    being used as ``assert recorded()`` for nicer readability.

    """
    if os.environ.get('MARV_TESTING_RECORD'):
        path.write_text(json.dumps(data, sort_keys=True, indent=2))
    recorded_data = json.loads(path.read_text())
    assert recorded_data == data, path
    return True


def scanner(directory, subdirs, filenames):  # pylint: disable=unused-argument
    root = Path(directory).name
    return [
        (f'{root}_{x}_filtér' if x == '0043' else f'{root}_{x}', [x])
        for x in filenames
    ]


@marv.node(Int8Value)
@marv.input('dataset', default=dataset_node)
def node_test(dataset):
    dataset = yield marv.pull(dataset)
    filename = Path(dataset.files[0].path).name
    yield marv.push({'value': int(filename)})


@marv.node(Words)
@marv.input('dataset', default=dataset_node)
def node_divisors(dataset):
    dataset = yield marv.pull(dataset)
    filename = Path(dataset.files[0].path).name
    idx = int(filename)
    divisors = [
        x for x in range(1, idx + 1)
        if idx % x == 0
    ]
    yield marv.push({'words': [f'div{x}' for x in divisors]})


@marv.node(Section)
@marv.input('node', default=node_test)
def section_test(node):
    value = yield marv.pull(node)
    value = value.value
    yield marv.push({'title': 'Test', 'widgets': [
        {'keyval': {'items': [{'title': 'value', 'cell': {'uint64': value}}]}},
    ]})


@pytest.fixture(scope='function')
async def site(loop, request, tmpdir):  # pylint: disable=unused-argument
    # pylint: disable=too-many-locals
    collection_names = ('hodge', 'podge')
    add_nodes = '\n'.join([
        '    marv.tests.conftest:node_test',
        '    marv.tests.conftest:section_test',
    ])
    add_filters = '    node_test | Test node | any | subset | (list (get "node_test.value" 0))'
    add_columns = '    node_test  | Test   | int      | (get "node_test.value" 0)'
    add_sections = '    section_test'
    collections = ''.join([
        COLLECTION_CONF.format(name='hodge',
                               add_nodes=add_nodes,
                               add_filters=add_filters,
                               add_columns=add_columns,
                               add_sections=add_sections),
        COLLECTION_CONF.format(name='podge',
                               add_nodes='',
                               add_filters='',
                               add_columns='',
                               add_sections=''),
    ])
    marv_conf = (tmpdir / 'marv.conf')
    mark = {x.name: x.kwargs for x in request.node.iter_markers()}
    site_cfg = mark.get('marv', {}).get('site', {})
    marv_conf.write(MARV_CONF.format(acl_name=site_cfg.get('acl', 'marv_webapi.acls:public'),
                                     collection_names=' '.join(collection_names),
                                     collections=collections))

    def create_datemock():
        vanilla_datetime = datetime.datetime
        side_effect = count(6000)

        class DatetimeMeta(type):
            @classmethod
            def __instancecheck__(cls, obj):
                return isinstance(obj, vanilla_datetime)

        class DatetimeBase(vanilla_datetime):
            @classmethod
            def utcnow(cls):
                ret = vanilla_datetime.fromtimestamp(next(side_effect))
                return ret

        return DatetimeMeta('datetime', (DatetimeBase,), {})

    def walker(path):
        collection = Path(path).name
        assert collection in collection_names, 'TEST SETUP ERROR: basename must match a collection!'
        idx = collection_names.index(collection)
        size = site_cfg.get('size', 50)
        return ((path, [], [f'{x+1:04d}' for x in range((idx + 1) * size)]),)

    def stat(path):
        filename = Path(path).name
        idx = int(filename)

        class Stat:  # pylint: disable=too-few-public-methods
            st_mtime = 1500000000 + idx * 1000
            st_size = idx
        return Stat()

    cls = site_cfg.get('cls', Site)
    site = await cls.create(str(marv_conf), init=True)  # pylint: disable=redefined-outer-name
    try:
        if not site_cfg.get('empty'):
            with mock.patch('bcrypt.gensalt', return_value=b'$2b$12$k67acf6S32i3nW0c7ycwe.'), \
                    mock.patch.object(datetime, 'datetime', create_datemock()), \
                    mock.patch('marv_node.setid.SetID.random', side_effect=count(2**127)), \
                    mock.patch('marv.utils.mtime', side_effect=count(1200000000)), \
                    mock.patch('marv.utils.stat', wraps=stat), \
                    mock.patch('marv.utils.now', side_effect=count(1400000000, 12 * 3600)), \
                    mock.patch('marv.utils.walk', wraps=walker), \
                    mock.patch('os.path.isdir', return_value=True), \
                    mock.patch('secrets.choice', return_value='#'):
                await site.db.user_add('test', password='test_pw', realm='marv', realmuid='',
                                       time_created=0xdead0001, time_updated=0xdead0002)
                await site.db.user_add('adm', password='adm_pw', realm='marv', realmuid='',
                                       time_created=0xdead0003, time_updated=0xdead0004)
                await site.db.group_adduser(groupname='admin', username='adm')

                prescan = site_cfg.get('prescan', None)
                if prescan:
                    await prescan(site)
                await site.scan()
                postscan = site_cfg.get('postscan', None)
                if postscan:
                    await postscan(site)

        yield site
    finally:
        await site.destroy()


@pytest.fixture(scope='function')
def app(site, request):  # pylint: disable=redefined-outer-name
    mark = {x.name: x.kwargs for x in request.node.iter_markers()}
    app_cfg = mark.get('marv', {}).get('app', {})
    cls = app_cfg.get('cls', App)
    app_ = cls(site).aioapp
    yield app_


@pytest.fixture(scope='function')
async def client(aiohttp_client, app):  # pylint: disable=redefined-outer-name
    client = await aiohttp_client(app)  # pylint: disable=redefined-outer-name

    headers = {
        'Content-Type': 'application/json',
    }

    async def authenticate(username, password):
        if 'Authorization' in headers:
            del headers['Authorization']
        resp = await client.post_json('/marv/api/auth', json={
            'username': username,
            'password': password,
        })
        headers['Authorization'] = f'Bearer {resp["access_token"]}'

    async def unauthenticate():
        return headers.pop('Authorization', '')

    async def get_json(*args, **kw):
        resp = await client.get(*args, headers=headers, **kw)
        if resp.status < 300:
            return await resp.json()
        return resp

    async def post_json(*args, **kw):
        resp = await client.post(*args, headers=headers, **kw)
        if resp.status < 300:
            return await resp.json()
        return resp

    client.headers = headers
    client.authenticate = authenticate
    client.get_json = get_json
    client.post_json = post_json
    client.unauthenticate = unauthenticate
    yield client



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only


async def test_rpc_query(site, client):
    # pylint: disable=too-many-statements

    sets = await site.db.get_datasets_for_collections(None)
    for setid in sets:
        await site.run(setid)

    res = await client.post('/marv/api/v1/rpcs', json={'rpcs': []})
    assert res.status == 401

    await client.authenticate('test', 'test_pw')

    # empty query
    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': []})
    assert res == {'data': {}}

    # most basic query
    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': [{'query': {
        'model': 'dataset',
    }}]})
    assert 'dataset' in res['data']
    assert len(res['data']['dataset']) == 150
    assert res['data']['dataset'][0] == {
        'collection': 'hodge',
        'discarded': 0,
        'id': 1,
        'name': 'hodge_0001',
        'setid': 'aaaaaaaaaaaaaaaaaaaaaaaaqa',
        'status': 0,
        'time_added': 1400000000000,
        'timestamp': 1200000000000,
    }

    # query with filtered attrs
    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': [{'query': {
        'model': 'dataset',
        'attrs': {
            'name': True,
        },
    }}]})
    assert res['data']['dataset'][0] == {'id': 1, 'name': 'hodge_0001'}

    # query filters
    def make_query(**query):
        query.setdefault('model', 'dataset')
        query.setdefault('attrs', {
            'collection': True,
            'name': True,
        })
        return {'rpcs': [{'query': query}]}

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'eq', 'name': 'id', 'value': 2},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
        ],
    }

    # check non ascii values
    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'eq', 'name': 'name', 'value': 'filtér'},
    ]))
    assert res['data'] == {'dataset': []}

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'substring', 'name': 'name', 'value': 'filtér'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 43, 'name': 'hodge_0043_filtér'},
            {'collection': 'podge', 'id': 93, 'name': 'podge_0043_filtér'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'substring', 'name': 'name', 'value': 'filt\xe9r'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 43, 'name': 'hodge_0043_filtér'},
            {'collection': 'podge', 'id': 93, 'name': 'podge_0043_filtér'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'lte', 'name': 'id', 'value': 2},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 1, 'name': 'hodge_0001'},
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'lt', 'name': 'id', 'value': 2},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 1, 'name': 'hodge_0001'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'gte', 'name': 'id', 'value': 149},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'podge', 'id': 149, 'name': 'podge_0099'},
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'gt', 'name': 'id', 'value': 149},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'between', 'name': 'id', 'value': [2, 4]},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'hodge', 'id': 3, 'name': 'hodge_0003'},
            {'collection': 'hodge', 'id': 4, 'name': 'hodge_0004'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'notbetween', 'name': 'id', 'value': [2, 149]},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 1, 'name': 'hodge_0001'},
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'in', 'name': 'id', 'value': [2, 4]},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'hodge', 'id': 4, 'name': 'hodge_0004'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'between', 'name': 'id', 'value': [2, 4]},
        {'op': 'ne', 'name': 'name', 'value': 'hodge_0003'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'hodge', 'id': 4, 'name': 'hodge_0004'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'between', 'name': 'id', 'value': [2, 5]},
        {'op': 'notin', 'name': 'name', 'value': ['hodge_0003', 'hodge_0005']},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'hodge', 'id': 4, 'name': 'hodge_0004'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'startswith', 'name': 'name', 'value': 'podge_01'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'startswith', 'name': 'name', 'value': '_'},
    ]))
    assert res['data'] == {'dataset': []}

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'endswith', 'name': 'name', 'value': '_0001'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 1, 'name': 'hodge_0001'},
            {'collection': 'podge', 'id': 51, 'name': 'podge_0001'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'endswith', 'name': 'name', 'value': '_'},
    ]))
    assert res['data'] == {'dataset': []}

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'substring', 'name': 'name', 'value': '_'},
    ]))
    assert len(res['data']['dataset']) == 150

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'substring', 'name': 'name', 'value': '_01'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'like', 'name': 'name', 'value': 'hodge'},
    ]))
    assert not res['data']['dataset']

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'like', 'name': 'name', 'value': 'hodge%'},
    ]))
    assert len(res['data']['dataset']) == 50

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'notlike', 'name': 'name', 'value': 'hodge%'},
    ]))
    assert len(res['data']['dataset']) == 100

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'isnot', 'name': 'name', 'value': None},
    ]))
    assert len(res['data']['dataset']) == 150

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'is', 'name': 'name', 'value': None},
    ]))
    assert not res['data']['dataset']

    # explict boolean logic
    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'not', 'value': {'op': 'eq', 'name': 'id', 'value': 2}},
    ]))
    assert len(res['data']['dataset']) == 149
    assert 2 not in [x['id'] for x in res['data']['dataset']]

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {
            'op': 'and',
            'value': [
                {'op': 'gt', 'name': 'id', 'value': 2},
                {'op': 'lt', 'name': 'id', 'value': 5},
            ],
        },
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 3, 'name': 'hodge_0003'},
            {'collection': 'hodge', 'id': 4, 'name': 'hodge_0004'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {
            'op': 'or',
            'value': [
                {'op': 'eq', 'name': 'id', 'value': 2},
                {'op': 'eq', 'name': 'name', 'value': 'podge_0001'},
            ],
        },
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'podge', 'id': 51, 'name': 'podge_0001'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'endswith', 'name': 'files.path', 'value': '0042'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 42, 'name': 'hodge_0042'},
            {'collection': 'podge', 'id': 92, 'name': 'podge_0042'},
        ],
    }

    res = await site.db.bulk_tag((
        ('important', 2),
        ('important', 3),
        ('important', 150),
        ('check', 2),
    ), (), '::')

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(filters=[
        {'op': 'eq', 'name': 'tags.value', 'value': 'important'},
    ]))
    assert res['data'] == {
        'dataset': [
            {'collection': 'hodge', 'id': 2, 'name': 'hodge_0002'},
            {'collection': 'hodge', 'id': 3, 'name': 'hodge_0003'},
            {'collection': 'podge', 'id': 150, 'name': 'podge_0100'},
        ],
    }

    # relation embedding
    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(
        filters=[{'op': 'eq', 'name': 'tags.value', 'value': 'important'}],
        attrs={
            'name': True,
            'tags': {'value': True},
        },
    ))
    assert res['data'] == {
        'dataset': [
            {'id': 2, 'name': 'hodge_0002', 'tags': [1, 2]},
            {'id': 3, 'name': 'hodge_0003', 'tags': [2]},
            {'id': 150, 'name': 'podge_0100', 'tags': [2]},
        ],
        'tag': [
            {'id': 1, 'value': 'check'},
            {'id': 2, 'value': 'important'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(
        filters=[{'op': 'eq', 'name': 'tags.value', 'value': 'important'}],
        attrs={
            'name': True,
            'files': True,
        },
    ))
    assert res['data']['dataset'] == [
        {'id': 2, 'name': 'hodge_0002', 'files': [2]},
        {'id': 3, 'name': 'hodge_0003', 'files': [3]},
        {'id': 150, 'name': 'podge_0100', 'files': [150]},
    ]
    assert len(res['data']['file']) == 3
    assert len(list(res['data']['file'][0].keys())) == 7

    # order, limit, offset
    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(
        attrs={
            'name': True,
        },
        order=['name', 'ASC'],
        limit=3,
        offset=3,
    ))
    assert res['data'] == {
        'dataset': [
            {'id': 4, 'name': 'hodge_0004'},
            {'id': 5, 'name': 'hodge_0005'},
            {'id': 6, 'name': 'hodge_0006'},
        ],
    }

    res = await client.post_json('/marv/api/v1/rpcs', json=make_query(
        attrs={
            'name': True,
        },
        order=['name', 'DESC'],
        limit=3,
        offset=3,
    ))
    assert res['data'] == {
        'dataset': [
            {'id': 147, 'name': 'podge_0097'},
            {'id': 146, 'name': 'podge_0096'},
            {'id': 145, 'name': 'podge_0095'},
        ],
    }

    # check users do not return password
    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': [{'query': {
        'model': 'user',
    }}]})
    assert 'user' in res['data']
    assert len(res['data']['user']) == 2
    for user in res['data']['user']:
        assert 'password' not in user

    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': [{'query': {
        'model': 'group',
        'attrs': {
            'users': True,
        },
    }}]})
    assert 'user' in res['data']
    assert [x['name'] for x in res['data']['user']] == ['adm']
    for user in res['data']['user']:
        assert 'password' not in user

    # internal users/groups are not returned
    res = await client.post_json('/marv/api/v1/rpcs', json={'rpcs': [{'query': {
        'model': 'user',
        'attrs': {
            'name': True,
            'groups': True,
        },
    }}]})
    assert res['data'] == {
        'group': [{'id': 3, 'name': 'admin'}],
        'user': [
            {'groups': [3], 'id': 3, 'name': 'adm'},
            {'groups': [], 'id': 2, 'name': 'test'},
        ],
    }



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from ..testing import make_dataset, marv, run_nodes


@marv.node()
def source():
    yield marv.push(1)
    yield marv.push(2)
    yield marv.push(3)


@marv.node()
@marv.input('stream', default=source)
def cubic(stream):
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        yield marv.push(msg**3)


DATASET = make_dataset()


async def test():
    nodes = [cubic]
    streams = await run_nodes(DATASET, nodes)
    assert streams == [
        [1, 8, 27],
    ]



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from abc import ABCMeta, abstractproperty
from logging import getLogger


class Keyed(metaclass=ABCMeta):

    @abstractproperty
    def key(self):
        return None  # pragma: nocoverage

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key < other.key

    def __le__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key <= other.key

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key == other.key

    def __ne__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key != other.key

    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key >= other.key

    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.key > other.key

    def __hash__(self):
        return hash((type(self), self.key))

    def __repr__(self):
        return f'<{type(self).__name__} key={self.key!r}>'


class AGenWrapperMixin:
    _agen = None

    @property
    def aclose(self):
        return self._agen.aclose

    @property
    def __anext__(self):
        return self._agen.__anext__

    @property
    def asend(self):
        return self._agen.asend

    @property
    def athrow(self):
        return self._agen.athrow


class GenWrapperMixin:
    _gen = None

    @property
    def close(self):
        return self._gen.close

    @property
    def __next__(self):
        return self._gen.__next__

    @property
    def send(self):
        return self._gen.send

    @property
    def throw(self):
        return self._gen.throw


class LoggerMixin:
    @property
    def logdebug(self):
        return self.log.debug

    @property
    def lognoisy(self):
        return self.log.noisy

    @property
    def logverbose(self):
        return self.log.verbose

    @property
    def loginfo(self):
        return self.log.info

    @property
    def logwarn(self):
        return self.log.warn

    @property
    def logerror(self):
        return self.log.error

    @property
    def log(self):
        logkey = ('marv', type(self).__name__.lower())
        if hasattr(self, 'key'):
            logkey += ((str(self.setid.abbrev),
                        self.node.name,
                        self.node.specs_hash[:10])
                       + tuple(str(x) for x in self.key[2:]))
        return getLogger('.'.join(logkey))


class Task(metaclass=ABCMeta):  # pylint: disable=too-few-public-methods
    __slots__ = ()

    def __repr__(self):
        return type(self).__name__


class Request(metaclass=ABCMeta):  # pylint: disable=too-few-public-methods
    __slots__ = ()

    def __repr__(self):
        return type(self).__name__



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import fcntl
import os
import shutil
import sqlite3
import sys
import sysconfig
from itertools import count, product
from logging import getLogger
from pathlib import Path
from uuid import uuid4

from pkg_resources import resource_filename
from pypika import SQLLiteQuery as Query, Table

from marv_node.run import run_nodes
from marv_store import Store
from . import utils
from .collection import Collections
from .config import Config
from .db import DBNotInitialized, DBVersionError
from .db import Database, Tortoise, create_or_ignore, scoped_session
from .model import Dataset, Group, User


DEFAULT_NODES = """
marv_nodes:dataset
marv_nodes:meta_table
marv_nodes:summary_keyval
"""

DEFAULT_FILTERS = """
name       | Name       | substring         | string   | (get "dataset.name")
setid      | Set Id     | startswith        | string   | (get "dataset.id")
size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
status     | Status     | any all           | subset   | (status)
tags       | Tags       | any all           | subset   | (tags)
comments   | Comments   | substring         | string   | (comments)
files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
"""

DEFAULT_LISTING_COLUMNS = """
name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
size       | Size   | filesize | (sum (get "dataset.files[:].size"))
status     | Status | icon[]   | (status)
tags       | Tags   | pill[]   | (tags)
time_added | Added  | datetime | (get "dataset.time_added")
"""

DEFAULT_LISTING_SUMMARY = """
datasets | datasets | int       | (len (rows))
size     | size     | filesize  | (sum (rows "size" 0))
"""

DEFAULT_DETAIL_SUMMARY_WIDGETS = """
summary_keyval
meta_table
"""

log = getLogger(__name__)


class UnknownNode(Exception):
    pass


def make_config(siteconf):
    sitedir = Path(siteconf).parent
    py_ver = sysconfig.get_python_version()
    defaults = {
        'marv': {
            'acl': 'marv_webapi.acls:authenticated',
            'dburi': 'sqlite://' + str(sitedir / 'db' / 'db.sqlite'),
            'frontenddir': str(sitedir / 'frontend'),
            'oauth': '',
            'reverse_proxy': None,
            'sessionkey_file': str(sitedir / 'sessionkey'),
            'staticdir': str(sitedir / resource_filename('marv_ludwig', 'static')),
            'storedir': str(sitedir / 'store'),
            'upload_checkpoint_commands': '',
            'venv': str(sitedir / 'venv'),
            'sitepackages': str(sitedir / 'venv' / 'lib' / f'python{py_ver}' / 'site-packages'),
            'window_title': '',
        },
        'collection': {
            'compare': None,
            'detail_summary_widgets': DEFAULT_DETAIL_SUMMARY_WIDGETS,
            'detail_sections': '',
            'detail_title': '(get "dataset.name")',
            'filters': DEFAULT_FILTERS,
            'listing_columns': DEFAULT_LISTING_COLUMNS,
            'listing_sort': '| ascending',
            'listing_summary': DEFAULT_LISTING_SUMMARY,
            'nodes': DEFAULT_NODES,
        },
    }
    required = {
        'marv': [
            'collections',
        ],
        'collection': [
            'scanner',
            'scanroots',
        ],
    }
    schema = {
        'marv': {
            'acl': 'find_obj',
            'collections': 'space_separated_list',
            'oauth': 'lines',
            'staticdir': 'path',
            'storedir': 'path',
            'upload_checkpoint_commands': 'cmd_lines',
        },
        'collection': {
            'compare': 'find_obj',
            'detail_title': 'function',
            'detail_summary_widgets': 'lines',
            'detail_sections': 'lines',
            'filters': 'lines',
            'listing_columns': 'lines',
            'listing_sort': 'pipe_separated_list',
            'listing_summary': 'lines',
            'nodes': 'lines',
            'scanroots': 'path_lines',
        },
    }
    return Config.from_file(siteconf, defaults=defaults,
                            required=required, schema=schema)


def load_sitepackages(sitepackages):
    import site  # pylint: disable=import-outside-toplevel
    site.USER_SITE = sitepackages
    sitepackages = Path(sitepackages)
    sitepackages.mkdir(parents=True, exist_ok=True)
    if str(sitepackages) not in sys.path:
        sys.path.append(str(sitepackages))
    try:
        with (sitepackages / 'easy-install.pth').open('r') as f:
            for directory in f.readlines():
                if directory[0] == '#' or directory.startswith('import'):
                    continue
                directory = directory.strip()
                if directory not in sys.path:
                    sys.path.append(directory)
    except FileNotFoundError:
        pass


class Site:
    Database = Database

    def __init__(self, siteconf):
        self.config = make_config(siteconf)
        # TODO: maybe ordereddict for meta, or generate multiple keys
        self.config.marv.oauth = {
            line.split(' ', 1)[0]: [field.strip() for field in line.split('|')]
            for line in self.config.marv.oauth
        }
        self.collections = Collections(config=self.config, site=self)
        self.db = self.Database()  # pylint: disable=invalid-name
        self.createdb = True

    @classmethod
    async def create(cls, siteconf, init=None):  # noqa: C901
        site = cls(siteconf)
        if utils.within_pyinstaller_bundle():
            load_sitepackages(site.config.marv.sitepackages)

        assert site.config.marv.dburi.startswith('sqlite://')
        dbpath = Path(site.config.marv.dburi.split('://', 1)[1])
        site.createdb = not dbpath.exists()
        if not site.createdb:
            site.check_db_version()
        elif not init:
            raise DBNotInitialized('There is no marv database.')

        if init:
            site.init_directory()

        try:
            fd = os.open(site.config.marv.sessionkey_file,
                         os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o600)
        except OSError as e:
            if e.errno != 17:
                raise
        else:
            with os.fdopen(fd, 'w') as f:
                f.write(str(uuid4()))
            log.verbose('Generated %s', site.config.marv.sessionkey_file)

        # Generate all dynamic models
        models = site.db.MODELS + [y for x in site.collections.values() for y in x.model]

        await Tortoise.init(config={
            'connections': {
                'default': {
                    'engine': 'tortoise.backends.sqlite',
                    'credentials': {
                        'file_path': dbpath,
                        'foreign_keys': 1,
                    },
                },
            },
            'apps': {
                'models': {
                    'models': models,
                },
            },
        })

        await site.db.initialize_connections()

        try:
            if init:
                await site.init_database()

            async with scoped_session(site.db) as txn:
                try:
                    await txn.execute_query('SELECT name FROM sqlite_master WHERE type="table"')
                except ValueError:
                    raise DBNotInitialized
        except BaseException:
            await site.destroy()
            raise

        return site

    async def destroy(self):
        await self.db.close_connections()
        await Tortoise.close_connections()

    def load_for_web(self):
        _ = [getattr(x, y) and None for x, y, in product(
            self.collections.values(),
            ('compare', 'filter_specs', 'listing_columns', 'model',
             'sortcolumn', 'sortorder', 'summary_items'))]

    def init_directory(self):
        try:
            os.mkdir(self.config.marv.storedir)
            log.verbose('Created %s', self.config.marv.storedir)
        except OSError as e:
            if e.errno != 17:
                raise

        try:
            dbdir = os.path.dirname(self.config.marv.dburi.replace('sqlite://', ''))
            os.mkdir(dbdir)
            log.verbose('Created %s', dbdir)
        except OSError as e:
            if e.errno != 17:
                raise

    def check_db_version(self):
        dbpath = Path(self.config.marv.dburi.split('://', 1)[1])
        metadata = Table('metadata')
        required = self.db.VERSION
        try:
            conn = sqlite3.connect(f'file:{dbpath}?mode=ro', uri=True)
            with conn:
                have = conn.execute(Query.from_(metadata)
                                    .select(metadata.value)
                                    .where(metadata.key == 'database_version')
                                    .get_sql())\
                           .fetchone()[0]
                if have != required:
                    raise DBVersionError(
                        f'DB version on disk "{have}", but "{required}" required.',
                    )
        except (sqlite3.OperationalError, TypeError):
            raise DBVersionError(
                'DB version is unknown, metadata table or version entry missing.',
            )
        finally:
            conn.close()

    async def store_db_version(self, txn):
        metadata = Table('metadata')
        await txn.exq(Query.into(metadata)
                      .columns(metadata.key, metadata.value)
                      .insert('database_version', self.db.VERSION))

    async def drop_listings(self, txn):
        prefixes = [f'l_{col}' for col in self.collections.keys()]
        tables = {
            name for name in [
                x['name'] for x in (await txn.execute_query(
                    'SELECT name FROM sqlite_master WHERE type="table"',
                ))[1]
            ]
            if any(name.startswith(prefix) for prefix in prefixes)
        }
        for table in sorted(tables, key=len, reverse=True):
            await txn.execute_query(f'DROP TABLE {table};')

    async def init_database(self):
        async with scoped_session(self.db) as txn:
            await self.drop_listings(txn)

        await Tortoise.generate_schemas()

        async with scoped_session(self.db) as txn:
            for name in ('marv:user:anonymous', 'marv:users', 'admin'):
                await Group.get_or_create(name=name, using_db=txn)

            for name in ('marv:anonymous',):
                user = await User.get_or_create(name=name, realm='marv', realmuid='',
                                                using_db=txn)
                await user[0].groups.add(
                    await Group.get(name=name.replace(':', ':user:')).using_db(txn),
                    using_db=txn,
                )

            if self.createdb:
                await self.store_db_version(txn)

            await create_or_ignore('acn', id=1, txn=txn)
            for name in self.collections:
                await create_or_ignore('collection', name=name, acn_id=1, txn=txn)

            log.verbose('Initialized database %s', self.config.marv.dburi)
            for col, collection in self.collections.items():
                loop = count()
                batchsize = 50
                # TODO: increase verbosity and probably introduce --reinit
                while True:
                    batch = await Dataset.filter(collection__name=col)\
                                         .using_db(txn)\
                                         .prefetch_related('files')\
                                         .limit(batchsize)\
                                         .offset(batchsize*next(loop))\
                                         .all()
                    for dataset in batch:
                        collection.render_detail(dataset)
                    if batch:
                        await collection.update_listings(batch, txn=txn)
                    if len(batch) < batchsize:
                        break
        log.info('Initialized from %s', self.config.filename)

    async def cleanup_discarded(self):
        descs = {
            key: x.table_descriptors
            for key, x in self.collections.items()
        }
        await self.db.cleanup_discarded(descs)
        # TODO: Cleanup corresponding store paths

    async def cleanup_relations(self):
        descs = {key: x.table_descriptors for key, x in self.collections.items()}
        await self.db.delete_listing_rel_values_without_ref(descs)

    async def restore_database(self, **kw):
        await self.db.restore_database(self, kw)

    async def run(self, setid, selected_nodes=None, deps=None, force=None, keep=None,
                  force_dependent=None, update_detail=None, update_listing=None,
                  excluded_nodes=None, cachesize=None):
        # pylint: disable=too-many-arguments,too-many-locals,too-many-branches

        assert not force_dependent or selected_nodes

        excluded_nodes = set(excluded_nodes or [])
        async with scoped_session(self.db) as txn:
            dataset = await Dataset.get(setid=setid)\
                                   .prefetch_related('collection', 'files')\
                                   .using_db(txn)
        collection = self.collections[dataset.collection.name]
        selected_nodes = set(selected_nodes or [])
        if not (selected_nodes or update_listing or update_detail):
            selected_nodes.update(collection.listing_deps)
            selected_nodes.update(collection.detail_deps)
        persistent = collection.nodes
        try:
            nodes = {persistent[name] if ':' not in name else utils.find_obj(name)
                     for name in selected_nodes
                     if name not in excluded_nodes
                     if name != 'dataset'}
        except KeyError as e:
            raise UnknownNode(dataset.collection, e.args[0])

        if force_dependent:
            nodes.update(x for name in selected_nodes for x in persistent[name].dependent)
        nodes = sorted(nodes)

        storedir = self.config.marv.storedir
        store = Store(storedir, persistent)

        changed = False
        try:
            if nodes:
                changed = await run_nodes(dataset, nodes, store, force=force,
                                          persistent=persistent,
                                          deps=deps, cachesize=cachesize)
        finally:
            if not keep:
                for stream in store.pending:
                    if stream.streamfile:
                        stream.streamfile.close()
                for stream in store.readstreams:
                    if stream.streamfile:
                        stream.streamfile.close()
                for tmpdir, tmpdir_fd in store.pending.values():
                    store.logdebug('Cleaning up %r', tmpdir)
                    shutil.rmtree(tmpdir)
                    fcntl.flock(tmpdir_fd, fcntl.LOCK_UN)
                    os.close(tmpdir_fd)
                store.pending.clear()

        if changed or update_detail:
            collection.render_detail(dataset)
            log.verbose('%s detail rendered', setid)
        if changed or update_listing:
            await collection.update_listings([dataset])
            log.verbose('%s listing rendered', setid)

        return changed

    async def scan(self, dry_run=None):
        for collection in self.collections.values():
            for scanroot in collection.scanroots:
                await collection.scan(scanroot, dry_run)



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import logging

from testfixtures import LogCapture

from ..testing import make_dataset, marv, run_nodes

DATASET = make_dataset()


@marv.node(group=True)
def source():
    yield marv.create_stream('Output 1')
    yield marv.create_stream('Output 2', foo=1)


@marv.node()
@marv.input('source', source)
def consumer(source):  # pylint: disable=redefined-outer-name
    logger = yield marv.get_logger()
    stream = yield marv.pull(source)
    logger.critical((stream.name, stream.header))
    stream = yield marv.pull(source)
    logger.critical((stream.name, stream.header))


async def test():
    with LogCapture(level=logging.CRITICAL) as log:
        await run_nodes(DATASET, [consumer])

    assert [x.msg for x in log.records if x.msg] == [
        ('Output 1', {}),
        ('Output 2', {'foo': 1}),
    ]



# Copyright 2016 - 2020  Ternaris, all rights reserved.
# SPDX-License-Identifier: AGPL-3.0-only

import math

import pytest

from marv.sexp import Identifier, List, Literal, scan
from marv.sexp import InvalidNumber, InvalidStringEscape, SexpError, UnexpectedCharacter
from marv.sexp import RESERVED_WORDS, ReservedWord


def test_parse_malformed():  # pylint: disable=too-many-statements
    sexp = ''
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ("Expected '('", sexp, 0)

    sexp = ' '
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ("Expected '('", sexp, 0)

    sexp = 'malformed'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ("Expected '('", sexp, 0)

    sexp = '('
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Unbalanced parentheses', sexp, -1)

    sexp = r'("\q")'
    with pytest.raises(InvalidStringEscape) as einfo:
        scan(sexp)
    assert einfo.value.args == (r'\q', sexp, 2)

    sexp = r'("\u)'
    with pytest.raises(InvalidStringEscape) as einfo:
        scan(sexp)
    assert einfo.value.args == (r'\u)', sexp, 2)

    sexp = r'("\ux1234")'
    with pytest.raises(InvalidStringEscape) as einfo:
        scan(sexp)
    assert einfo.value.args == (r'\ux123', sexp, 2)

    sexp = '(-a)'
    with pytest.raises(InvalidNumber) as einfo:
        scan(sexp)
    assert einfo.value.args == ('-a', sexp, 1)

    sexp = '(")'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Runaway string', sexp, 1)

    sexp = '("foo)'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Runaway string', sexp, 1)

    sexp = '(()())'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Missing space between tokens', sexp, 3)

    sexp = '(() '
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Unbalanced parentheses', sexp, -1)

    sexp = '(a'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Unbalanced parentheses', sexp, -1)

    sexp = '(1'
    with pytest.raises(SexpError) as einfo:
        scan(sexp)
    assert einfo.value.args == ('Unbalanced parentheses', sexp, -1)

    sexp = '(+)'
    with pytest.raises(UnexpectedCharacter) as einfo:
        scan(sexp)
    assert einfo.value.args == ('+', sexp, 1)

    sexp = '(01)'
    with pytest.raises(InvalidNumber):
        scan(sexp)

    sexp = '(-01)'
    with pytest.raises(InvalidNumber):
        scan(sexp)

    sexp = '(1-)'
    with pytest.raises(Exception):
        scan(sexp)

    sexp = '(1\f\v\u2000\u2001)'
    rv = scan(sexp)
    assert rv.args[0].stop == 1  # pylint: disable=no-member

    sexp = '(foo-bar)'
    with pytest.raises(Exception):
        scan(sexp)


@pytest.mark.parametrize('string', range(0x20))
def test_invalid_character(string):
    sexp = f'("{chr(string)}")'
    with pytest.raises(Exception):
        scan(sexp)


@pytest.mark.parametrize('string', [
    '-inf', '-infinity', '-Inf', '-Infinity', '-nan', '-NaN',
])
def test_invalid_number(string):
    sexp = f'({string})'
    with pytest.raises(InvalidNumber):
        scan(sexp)


@pytest.mark.parametrize('string', RESERVED_WORDS)
def test_reserved_word(string):
    sexp = f'({string})'
    with pytest.raises(ReservedWord):
        scan(sexp)


def test_parse_empty():
    rv = scan('()')
    assert rv == List((), 0, 1)

    rv = scan('(())')
    assert rv == List((List((), 1, 2),), 0, 3)

    rv = scan('(() \t ()\r\n)')
    assert rv == List((List((), 1, 2), List((), 6, 7)), 0, 10)

    rv = scan('((()) ())')
    assert rv == List((List((List((), 2, 3),), 1, 4), List((), 6, 7)), 0, 8)


def test_parse_literals():
    """Test support for JSON literals."""
    rv = scan(r'("foo \" \u0022 \/ \\ \b \f \n \r \t )(" "")')
    assert rv == List((Literal('foo " " / \\ \b \f \n \r \t )(', 1, 39),
                       Literal('', 41, 42)), 0, 43)

    rv = scan('(null)')
    assert rv == List((Literal(None, 1, 4),), 0, 5)

    rv = scan('(true false)')
    assert rv == List((Literal(True, 1, 4), Literal(False, 6, 10)), 0, 11)

    rv = scan('(0 -1 42)')
    assert rv == List((Literal(0, 1, 1), Literal(-1, 3, 4), Literal(42, 6, 7)), 0, 8)
    assert rv.args[1].value == -1  # pylint: disable=no-member

    rv = scan('(1.2 -1e10 -1E+10 -1.42e-10)')
    assert rv == List((Literal(1.2, 1, 3), Literal(-1e10, 5, 9), Literal(-1e10, 11, 16),
                       Literal(-1.42e-10, 18, 26)), 0, 27)

    rv = scan('(-0.0)')
    minus_one = math.copysign(1, rv.args[0].value)  # pylint: disable=no-member
    assert minus_one == -1


def test_parse_symbols():
    rv = scan('(foo bar)')
    assert rv == List((Identifier('foo', 1, 3), Identifier('bar', 5, 7)), 0, 8)



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import json

import pytest

from marv.db import UnknownOperator


async def test_listing(site):
    # pylint: disable=too-many-statements

    sets = await site.db.get_datasets_for_collections(['hodge'])
    for setid in sets:
        await site.run(setid)

    await site.db.comment_by_setids([sets[2]], 'test', 'lorem ipsum')

    res = await site.db.bulk_tag([
        ('foo', 1),
        ('foo', 2),
        ('bar', 1),
    ], [], '::')

    collection = site.collections['hodge']
    descs = collection.table_descriptors

    res = await site.db.get_all_known_for_collection(site.collections, 'hodge', '::')
    assert res == {
        'divisors': [f'div{x}' for x in range(1, 51)],
        'node_test': [str(x) for x in range(1, 51)],
        'status': ['error', 'missing', 'outdated', 'pending'],
        'tags': ['bar', 'foo'],
    }

    res = await site.db.get_filtered_listing(descs, [], '::')
    assert len(res) == 50

    # by comment
    res = await site.db.get_filtered_listing(descs, [
        ('comments', 'rem', 'substring', 'string'),
    ], '::')
    assert len(res) == 1
    assert res[0]['id'] == 3

    # by tag
    res = await site.db.get_filtered_listing(descs, [
        ('tags', ['bar', 'foo'], 'any', 'string'),
    ], '::')
    assert len(res) == 2
    assert res[0]['id'] == 1
    assert res[1]['id'] == 2

    res = await site.db.get_filtered_listing(descs, [
        ('tags', ['bar', 'foo'], 'all', 'string'),
    ], '::')
    assert len(res) == 1
    assert res[0]['id'] == 1

    with pytest.raises(UnknownOperator):
        res = await site.db.get_filtered_listing(descs, [
            ('tags', ['bar', 'foo'], 'bad operator', 'string'),
        ], '::')

    # TODO: status

    # datetime
    day1 = 1400000000000 + 24 * 3600 * 1000
    day2 = day1 + 24 * 3600 * 1000
    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'eq', 'datetime'),
    ], '::')
    assert len(res) == 2
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(day1 <= x < day2 for x in times)

    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'ne', 'datetime'),
    ], '::')
    assert len(res) == 48
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(x < day1 or x >= day2 for x in times)

    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'lt', 'datetime'),
    ], '::')
    assert len(res) == 2
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(x < day1 for x in times)

    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'le', 'datetime'),
    ], '::')
    assert len(res) == 4
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(x < day2 for x in times)

    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'gt', 'datetime'),
    ], '::')
    assert len(res) == 46
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(x > day1 for x in times)

    res = await site.db.get_filtered_listing(descs, [
        ('time_added', day1, 'ge', 'datetime'),
    ], '::')
    assert len(res) == 48
    times = [json.loads(x['row'])['values'][4] for x in res]
    assert all(x >= day1 for x in times)

    # substring
    res = await site.db.get_filtered_listing(descs, [
        ('setid', 'eaa', 'substring', 'string'),
    ], '::')
    assert sorted(x['id'] for x in res) == [2, 10, 18, 26, 33, 34, 42, 50]

    # startswith
    res = await site.db.get_filtered_listing(descs, [
        ('setid', 'feaa', 'startswith', 'string'),
    ], '::')
    assert len(res) == 1
    assert res[0]['id'] == 42

    # lt
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'lt', 'int')], '::')
    assert len(res) == 9

    # le
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'le', 'int')], '::')
    assert len(res) == 10

    # eq
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'eq', 'int')], '::')
    assert len(res) == 1

    # ge
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'ge', 'int')], '::')
    assert len(res) == 41

    # gt
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'gt', 'int')], '::')
    assert len(res) == 40

    # ne
    res = await site.db.get_filtered_listing(descs, [('size', 10, 'ne', 'int')], '::')
    assert len(res) == 49

    # relations
    res = await site.db.get_filtered_listing(descs, [
        ('divisors', ['div17'], 'any', 'string'),
    ], '::')
    assert [x['id'] for x in res] == [17, 34]

    res = await site.db.get_filtered_listing(descs, [
        ('divisors', ['div2', 'div17'], 'all', 'string'),
    ], '::')
    assert [x['id'] for x in res] == [34]

    res = await site.db.get_filtered_listing(descs, [
        ('divisors', 'iv4', 'substring_any', 'string'),
    ], '::')
    assert [x['id'] for x in res] == list(range(4, 41, 4)) + list(range(41, 50))

    # TODO: words


async def test_listing_relations(site):
    sets = await site.db.get_datasets_for_collections(['hodge'])
    for setid in sets:
        await site.run(setid)

    res = await site.db.get_all_known_for_collection(site.collections, 'hodge', '::')
    assert res['divisors'] == [f'div{x}' for x in range(1, 51)]

    await site.db.discard_datasets_by_setids([sets[41]])
    await site.cleanup_discarded()

    res = await site.db.get_all_known_for_collection(site.collections, 'hodge', '::')
    assert res['divisors'] == [f'div{x}' for x in range(1, 51)]

    await site.cleanup_relations()
    res = await site.db.get_all_known_for_collection(site.collections, 'hodge', '::')
    assert res['divisors'] == [f'div{x}' for x in range(1, 51) if x != 42]



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from json import JSONDecodeError

from aiohttp import web

from marv.db import DBPermissionError
from .tooling import HTTPPermissionError, api_endpoint as marv_api_endpoint


@marv_api_endpoint('/dataset', methods=['DELETE'])
async def delete(request):
    try:
        ids = await request.json()
    except JSONDecodeError:
        raise web.HTTPBadRequest

    if not ids:
        raise web.HTTPBadRequest

    try:
        await request.app['site'].db.discard_datasets_by_dbids(ids, True, user=request['username'])
    except DBPermissionError:
        raise HTTPPermissionError(request)

    return web.json_response({})



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from json import JSONDecodeError

from aiohttp import web

from marv.db import DBPermissionError
from .tooling import HTTPPermissionError, api_endpoint as marv_api_endpoint


@marv_api_endpoint('/tag', methods=['POST'])  # noqa: C901
async def tag(request):  # noqa: C901
    try:
        changes = await request.json()
    except JSONDecodeError:
        raise web.HTTPBadRequest

    if not changes:
        raise web.HTTPBadRequest

    try:
        add = []
        remove = []
        for ops in changes.values():
            for opname, target in (('add', add), ('remove', remove)):
                for tagname, ids in ops.pop(opname, {}).items():
                    for id in ids:
                        target.append((tagname, id))
            if ops:
                raise web.HTTPBadRequest
    except AttributeError:
        raise web.HTTPBadRequest

    try:
        await request.app['site'].db.bulk_tag(add, remove, user=request['username'])
    except DBPermissionError:
        raise HTTPPermissionError(request)
    return web.json_response({})



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

from ..testing import make_dataset, marv, run_nodes


@marv.node()
def source():
    yield 1


DATASET = make_dataset()


async def test():
    nodes = [source]
    with pytest.raises(RuntimeError):  # TODO: proper exception
        await run_nodes(DATASET, nodes)



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import logging

from testfixtures import LogCapture

from ..testing import make_dataset, marv, run_nodes

DATASET = make_dataset()
SETID = DATASET.setid


@marv.node(group=True)
def source():
    out1 = yield marv.create_stream('Output 1', foo=1)
    out2 = yield marv.create_stream('Output 2', foo=2)
    out3 = yield marv.create_stream('Output 3', foo=3)
    yield out1.msg(a=1)
    yield out2.msg(b=10)
    yield out3.msg(c=100)
    yield out1.msg(a=2)
    yield out2.msg(b=20)
    yield out3.msg(c=200)


@marv.node()
@marv.input('stream', foreach=source)
def foreach(stream):
    logger = yield marv.get_logger()
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        logger.critical((0, stream.name, msg))


@marv.node()
@marv.input('node', default=source)
def withall(node):
    logger = yield marv.get_logger()
    streams = []
    while True:
        stream = yield marv.pull(node)
        if stream is None:
            break
        streams.append(stream)

    while streams:
        msgs = yield marv.pull_all(*streams)
        logger.critical((1, msgs))
        streams = [stream for stream, msg in zip(streams, msgs) if msg is not None]


async def test():
    with LogCapture(level=logging.CRITICAL) as log:
        await run_nodes(DATASET, [foreach, withall])

    assert [x.msg for x in log.records if x.msg[0] == 0] == [
        (0, 'Output 1', {'a': 1}),
        (0, 'Output 2', {'b': 10}),
        (0, 'Output 3', {'c': 100}),
        (0, 'Output 1', {'a': 2}),
        (0, 'Output 2', {'b': 20}),
        (0, 'Output 3', {'c': 200}),
    ]
    assert [x.msg for x in log.records if x.msg[0] == 1] == [
        (1, [{'a': 1}, {'b': 10}, {'c': 100}]),
        (1, [{'a': 2}, {'b': 20}, {'c': 200}]),
        (1, [None, None, None]),
    ]



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import math
import mimetypes
import time
from collections import OrderedDict
from pathlib import Path

import jwt
from aiohttp import web


async def check_authorization(request, acl, authorization):
    username = 'marv:anonymous'
    groups = {'__unauthenticated__'}
    if authorization:
        token = authorization.replace('Bearer ', '')
        try:
            session = jwt.decode(token, request.app['config']['SECRET_KEY'], algorithms=['HS256'])
        except BaseException:
            raise web.HTTPUnauthorized()

        user = await request.app['site'].db.get_user_by_name(session['sub'], deep=True)
        if not user or user.time_updated.timestamp() > session['iat']:
            raise web.HTTPUnauthorized()

        username = user.name
        groups = {g.name for g in user.groups}
        groups.add('__authenticated__')

    elif '__unauthenticated__' not in acl:
        raise web.HTTPUnauthorized()

    if acl and not acl.intersection(groups):
        raise web.HTTPForbidden()

    request['username'] = username
    request['user_groups'] = groups


def HTTPPermissionError(request):  # pylint: disable=invalid-name
    if request['username'] == 'marv:anonymous':
        return web.HTTPUnauthorized
    return web.HTTPForbidden


def get_global_granted(request):
    granted = {
        view.acl_key
        for name, view in request.app['api_endpoints'].items()
        if isinstance(view, APIEndpoint)
        if view.acl.intersection(request['user_groups'])
    }
    return sorted(granted - set(request.app['route_acl'].keys()))


def get_local_granted(request):
    return sorted({
        key
        for key, groups in request.app['route_acl'].items()
        if set(groups).intersection(request['user_groups'])
    })


def generate_token(username, key):
    now = math.ceil(time.time())
    return jwt.encode({
        'exp': now + 2419200,  # 4 weeks expiration
        'iat': now,
        'sub': username,
    }, key, algorithm='HS256')


def safejoin(basepath, rel):
    rel = Path(rel)
    if rel.anchor:
        raise web.HTTPForbidden

    fullpath = basepath.joinpath(rel).resolve()
    if basepath.resolve() not in fullpath.parents:
        raise web.HTTPForbidden

    return fullpath


def sendfile(path, approot, reverse_proxy, filename=None):
    headers = {
        'Cache-Control': 'no-cache',
        'Content-Disposition': f'attachment; filename={filename or path.name}',
    }

    if reverse_proxy == 'nginx':
        mime = mimetypes.guess_type(str(path))[0]
        return web.Response(headers={
            'Content-Type': mime or 'application/octet-stream',
            'X-Accel-Buffering': 'no',
            'X-Accel-Redirect': f'{approot}{str(path)}',
            **headers,
        })

    assert not reverse_proxy, f'Unknown reverse_proxy {reverse_proxy}'
    return web.FileResponse(path, headers=headers)


class APIEndpoint:
    MIN_API_VERSION = 3
    LATEST_API_VERSION = 3
    HEADER_PREFIX = 'application/vnd.marv.v'
    acl = None

    def __init__(self, name, func, url_rule, defaults=None, methods=None, version=None,
                 acl_key=None, force_acl=None):
        # pylint: disable=too-many-arguments
        self.name = name
        version = self.MIN_API_VERSION if version is None else version
        self.funcs = [(version, func)]
        self.url_rules = [(url_rule, {'defaults': defaults, 'methods': methods})]
        self.force_acl = set(force_acl) if force_acl else None
        self.acl_key = acl_key or name

    async def __call__(self, request):
        authorization = request.headers.get('Authorization')
        # TODO: can authorization be '' or is None test?
        if not authorization:
            authorization = request.query.get('access_token')
        await check_authorization(request, self.acl, authorization)

        try:
            accepted = next(x[0] for x in request.headers.getall('ACCEPT', [])
                            if x[0].startswith(self.HEADER_PREFIX))
            accepted_version = int(accepted[len(self.HEADER_PREFIX):])
        except (StopIteration, ValueError):
            accepted_version = self.MIN_API_VERSION

        try:
            func = next(func for version, func in self.funcs
                        if version <= accepted_version)
        except StopIteration:
            raise web.HTTPNotAcceptable()

        return await func(request)

    def init_app(self, app, url_prefix=None, name_prefix=None, app_root=None):
        self.acl = self.force_acl if self.force_acl else set(app['route_acl'][self.acl_key])
        name = '.'.join(filter(None, [name_prefix, self.name]))
        for url_rule, options in self.url_rules:
            url_rule = ''.join(filter(None, [url_prefix, url_rule]))
            if options['methods'] is None:
                options['methods'] = ['GET']
            assert len(options['methods']) == 1
            app['api_endpoints'][name] = self
            app.add_routes([web.route(options['methods'][0],
                                      f'{app_root}{url_rule}',
                                      self.__call__, name=name)])


class APIGroup:
    def __init__(self, name, func, url_prefix=None):
        self.name = name
        self.func = func
        self.url_prefix = url_prefix
        self.endpoints = OrderedDict()

    def add_endpoint(self, ep):
        """Endpoints and groups are all the same (for now)."""
        assert ep.name not in self.endpoints, ep
        self.endpoints[ep.name] = ep

    def endpoint(self, *args, **kw):
        return api_endpoint(*args, registry=self.endpoints, **kw)

    def init_app(self, app, url_prefix=None, name_prefix=None, app_root=None):
        self.func(app)
        name_prefix = '.'.join(filter(None, [name_prefix, self.name]))
        url_prefix = '/'.join(filter(None, [url_prefix, self.url_prefix])) or None
        for ep in self.endpoints.values():
            ep.init_app(app, url_prefix=url_prefix, name_prefix=name_prefix, app_root=app_root)

    def __repr__(self):
        return f'<APIGroup {self.name} url_prefix={self.url_prefix}>'


def api_endpoint(url_rule, defaults=None, methods=None, version=None,
                 cls=APIEndpoint, registry=None, acl_key=None, force_acl=None):
    # pylint: disable=too-many-arguments

    def decorator(func):
        if isinstance(func, cls):
            func.url_rules.append((url_rule, {'defaults': defaults, 'methods': methods}))
            return func

        name = func.__name__
        rv = cls(name, func, url_rule=url_rule, defaults=defaults,
                 methods=methods, version=version, acl_key=acl_key, force_acl=force_acl)
        rv.__doc__ = func.__doc__  # pylint: disable=attribute-defined-outside-init

        if registry is not None:
            assert name not in registry, name
            registry[name] = rv
        return rv
    return decorator


def api_group(url_prefix=None, cls=APIGroup):
    def decorator(func):
        if isinstance(func, cls):
            raise TypeError('Attempted to convert function into api group twice.')

        name = func.__name__
        rv = cls(name, func, url_prefix)
        rv.__doc__ = func.__doc__  # pylint: disable=attribute-defined-outside-init
        return rv
    return decorator



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from pkg_resources import iter_entry_points

from .auth import auth
from .collection import collection, meta
from .comment import comment
from .dataset import dataset as dataset_api
from .delete import delete
from .rpcs import rpcs
from .tag import tag
from .tooling import api_group as marv_api_group


@marv_api_group()
def webapi(_):
    pass


# Groups and endpoints are all the same for now
webapi.add_endpoint(auth)
webapi.add_endpoint(comment)
webapi.add_endpoint(dataset_api)
webapi.add_endpoint(delete)
webapi.add_endpoint(collection)
webapi.add_endpoint(meta)
webapi.add_endpoint(rpcs)
webapi.add_endpoint(tag)


def load_entry_points():
    for entry_point in iter_entry_points(group='marv_webapi'):
        endpoint = entry_point.load()
        webapi.add_endpoint(endpoint)



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import hashlib
from base64 import b32encode
from collections import OrderedDict, namedtuple
from itertools import count, product

from marv_api import dag
from marv_api.utils import find_obj
from . import io
from .io import get_logger, pull
from .mixins import Keyed

NODE_CACHE = {}


class InputSpec(Keyed, namedtuple('InputSpec', ('name', 'value', 'foreach'))):
    @property
    def key(self):
        value = self.value.key if hasattr(self.value, 'key') else self.value
        return (self.name, value, self.foreach)

    def clone(self, value):
        cls = type(self)
        value = StreamSpec(value) if isinstance(value, Node) else value
        return cls(self.name, value, self.foreach)

    def __repr__(self):
        foreach = 'foreach ' if self.foreach else ''
        return f'<{type(self)} {foreach}{self.name}={self.value!r}>'


class StreamSpec:  # pylint: disable=too-few-public-methods
    def __init__(self, node, name=None):  # pylint: disable=redefined-outer-name
        assert isinstance(node, Node)
        self.node = node
        self.name = name
        self.key = (node.key,) if name is None else (node.key, name)
        self.args = (node,) if name is None else (node, name)


class Node(Keyed):  # pylint: disable=too-many-instance-attributes
    _key = None
    group = None

    @property
    def key(self):
        key = self._key
        if key:
            return key
        return '-'.join([self.specs_hash, self.fullname])

    @property
    def abbrev(self):
        return '.'.join([self.name, self.specs_hash[:10]])

    @staticmethod
    def genhash(specs):
        spec_keys = repr(tuple(x.key for x in sorted(specs.values()))).encode('utf-8')
        return b32encode(hashlib.sha256(spec_keys).digest()).decode('utf-8').lower()[:-4]

    @classmethod
    def from_dag_node(cls, func):
        if hasattr(func, '__marv_node__'):
            dnode = func.__marv_node__
        else:
            dnode = func
            func = None

        node = NODE_CACHE.get(dnode)
        if node is not None:
            return node

        if func is None:
            func = find_obj(dnode.function)
        namespace, name = dnode.function.rsplit('.', 1)
        schema = find_obj(dnode.message_schema) if dnode.message_schema is not None else None
        inputs = dnode.inputs
        specs = OrderedDict((
            (name, InputSpec(name=name,
                             value=(value if not isinstance(value, dag.Stream) else
                                    StreamSpec(node=Node.from_dag_node(value.node),
                                               name=value.name)),
                             foreach=name == dnode.foreach))
            for name, value in ((name, getattr(inputs, name)) for name in inputs.__fields__.keys())
        ))
        node = cls(func,
                   schema=schema,
                   version=dnode.version,
                   name=name,
                   namespace=namespace,
                   specs=specs,
                   group=dnode.group,
                   dag_node=dnode)
        NODE_CACHE[dnode] = node
        return node

    def __init__(self, func, schema=None, version=None,
                 name=None, namespace=None, specs=None, group=None, dag_node=None):
        # pylint: disable=too-many-arguments
        # TODO: assert no default values on func, or consider default
        # values (instead of) marv.input() declarations
        # Definitely not for node inputs as they are not passed directly!
        self.func = func
        self.version = version
        self.name = name = func.__name__ if name is None else name
        self.namespace = namespace = func.__module__ if namespace is None else namespace
        self.fullname = (name if not namespace else ':'.join([namespace, name]))
        self.specs_hash = self.genhash(specs or {})
        self.schema = schema
        self.specs = specs or {}
        assert group in (None, False, True, 'ondemand'), group
        self.group = (group if group is not None else
                      any(x.foreach for x in self.specs.values()))
        # TODO: StreamSpec, seriously?
        self.deps = {x.value.node for x in self.specs.values()
                     if isinstance(x.value, StreamSpec)}
        self.alldeps = self.deps.copy()
        self.alldeps.update(x for dep in self.deps
                            for x in dep.alldeps)

        self.consumers = set()
        for dep in self.deps:
            assert self not in dep.consumers, (dep, self)
            dep.consumers.add(self)

        self.dependent = set()
        for dep in self.alldeps:
            assert self not in dep.dependent, (dep, self)
            dep.dependent.add(self)

        if self.fullname == 'marv_nodes:dataset':
            self.load = func.load

        assert dag_node is not None
        self.dag_node = dag_node

    def __call__(self, **inputs):
        return self.func(**inputs)

    async def invoke(self, inputs=None):  # noqa: C901
        # pylint: disable=too-many-locals,too-many-branches
        # We must not write any instance variables, a node is running
        # multiple times in parallel.

        common = []
        foreach_plain = []
        foreach_stream = []
        if inputs is None:
            for spec in self.specs.values():
                assert not isinstance(spec.value, Node), (self, spec.value)
                if isinstance(spec.value, StreamSpec):
                    value = yield io.GetStream(setid=None, node=spec.value.node,
                                               name=spec.value.name or 'default')
                    target = foreach_stream if spec.foreach else common
                else:
                    value = spec.value
                    target = foreach_plain if spec.foreach else common
                target.append((spec.name, value))

        if foreach_plain or foreach_stream:
            log = yield get_logger()
            cross = product(*[[(k, x) for x in v] for k, v in foreach_plain])
            if foreach_stream:
                assert len(foreach_stream) == 1, self  # FOR NOW
                cross = list(cross)
                name, stream = foreach_stream[0]
                idx = count()
                while True:
                    value = yield pull(stream)
                    if value is None:
                        log.noisy('finished forking')
                        break
                    for inputs in cross:  # pylint: disable=redefined-argument-from-local
                        inputs = dict(inputs)
                        inputs.update(common)
                        inputs[name] = value
                        # TODO: give fork a name
                        _idx = next(idx)
                        log.noisy('FORK %d with: %r', _idx, inputs)
                        yield io.Fork(f'{_idx}', inputs, False)
            else:
                # pylint: disable=redefined-argument-from-local
                for _idx, inputs in enumerate(cross):
                    # TODO: consider deepcopy
                    inputs = dict(inputs)
                    inputs.update(common)
                    log.noisy('FORK %d with: %r', _idx, inputs)
                    yield io.Fork(f'{_idx}', inputs, False)
                log.noisy('finished forking')
        else:
            if inputs is None:
                inputs = dict(common)
            gen = self.func(**inputs)
            assert hasattr(gen, 'send')
            response = None
            while True:
                try:
                    request = gen.send(response)
                except StopIteration:
                    return
                response = yield request

    def __str__(self):
        return self.key

    def __repr__(self):
        return f'<Node {self.abbrev}>'



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from marv_api import (
    DatasetInfo as _DatasetInfo,
    deprecation,
)

__all__ = ()
__dir__, __getattr__ = deprecation.dir_and_getattr(__name__, __all__, {
    'DatasetInfo': deprecation.Info(__name__, '20.07', _DatasetInfo),
})
